package edu.sdsu.cs.util;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.NoSuchElementException;

/*
 * Trie over an arbitrary alphabet
 */
public class Trie {
	TrieNode root;
	Alphabet characterSet;
	int modificationCount = 0;
	
	public Trie() {
		this(Alphabet.ascii());
	}
	
	public Trie(Alphabet characters) {
		characterSet = characters;
		root = new TrieInnerNode(characterSet);
	}
	
	public void accept(TrieVisitor visitor) {
		visitor.visitTrie(this);
		TrieNodeIterator nodes = new TrieNodeIterator(root);
		while (nodes.hasNext())
			nodes.next().accept(visitor);
	}
	
	public boolean add(String wordToAdd){
		if (wordToAdd == null) return false;
		if (wordToAdd.isEmpty()) return false;
		if (!characterSet.validCharacters(wordToAdd)) 
			throw new IllegalArgumentException(wordToAdd + " contains characters not in alphabet: " + characterSet);
		modificationCount++;
		return root.add(new WordParameter(wordToAdd.toLowerCase()));
	}
	
	public boolean contains(String word) {
		if (word == null) return false;
		if (word.isEmpty()) return false;
		if (!characterSet.validCharacters(word)) 
			throw new IllegalArgumentException(word + " contains characters not in alphabet: " + characterSet);
		return root.contains(new WordParameter(word.toLowerCase()));
	}
	
	public boolean isEmpty() {
		return size() == 0;
	}

	public Iterator<String> iterator() {
		return new TrieIterator(root);
	}
	
	public int size(){
		return root.size();
	}
	
	public String toString() {
		ArrayList<String> words = new ArrayList<String>();
		Iterator wordIterator = iterator();
		while (wordIterator.hasNext())
			words.add((String)wordIterator.next());
		return words.toString();
	}
	
	private class TrieIterator implements Iterator<String> {
		//Class invariant = always prefetch next element to simplify hasNext logic
		private TrieNodeIterator nodeIterator;
		private TrieWordNode next;
		int initialModificationCount;
		
		public TrieIterator(TrieNode root) {
			nodeIterator = new TrieNodeIterator(root);
			next = getNextWordNode();
			initialModificationCount = modificationCount;
		}
		
		@Override
		public boolean hasNext() {
			if (initialModificationCount != modificationCount) throw new UnsupportedOperationException();
			return next != null;
		}

		@Override
		public String next() {
			if (initialModificationCount != modificationCount) throw new UnsupportedOperationException();
			if (next == null) throw new NoSuchElementException ();
			TrieNode current = next;
			next = getNextWordNode();
			return current.getWord();
		}

		@Override
		public void remove() {
			throw new UnsupportedOperationException();
		}

		private TrieWordNode getNextWordNode() {
			if (!nodeIterator.hasNext()) return null;
			TrieNode nextNode = nodeIterator.next();
			while (nodeIterator.hasNext() && !nextNode.isWord()) {
				nextNode = nodeIterator.next();
			}
			if (!nextNode.isWord()) return null;
			return (TrieWordNode) nextNode;
		}
	}
}
