package edu.sdsu.cs.util;

import java.util.Arrays;

/*
 * Represents a node in a trie that is not a Word Node
 */
public class TrieInnerNode extends TrieNode {
	private static final char NULL_CHAR = '\n';
	protected TrieNode[] subtries;
	private char character;
	
	public TrieInnerNode() {
		this(Alphabet.ascii());
	}
	
	public TrieInnerNode(Alphabet letters) {
		this(letters, null);
	}
	
	public TrieInnerNode(Alphabet letters, TrieInnerNode parent) {
		this(letters, parent, NULL_CHAR);
	}
	
	public TrieInnerNode(Alphabet letters, TrieInnerNode parent, char character) {
		super(letters,parent);
		this.character = character;
		this.subtries = initialSubtries(letters);
	}
	
	public void accept(TrieVisitor visitor) {
		visitor.visitTrieInnerNode(this);
	}
	
	public boolean add(WordParameter wordToAdd) {
		if (wordToAdd.atEnd()) {
			becomeWordNode();
			return true;
		}
		char whereToAdd = wordToAdd.nextChar();
		return getNode(whereToAdd).add(wordToAdd.nextLevel());
	}

	@Override
	public boolean contains(WordParameter word) {
		if (word.atEnd()) return false;
		char nextChar = word.nextChar();
		return getNode(nextChar).contains(word.nextLevel());
	}
	
	@Override
	public String getWord() {
		if (isRootNode()) return "";
		return parent.getWord() + character;
	}
	
	@Override
	public int size() {
		int size = 0;
		for (TrieNode each: subtries) {
			size = size + each.size();
		}
		return size;
	}
	
	/*
	 * Replace current node with TrieWordNode object
	 */
	protected void becomeWordNode() {
		//Point all references to current node to myReplacement
		//WordNode constructor points all my subtrie parent references to myReplacement
		TrieWordNode myReplacement = new TrieWordNode(characterSet, parent, character, subtries);
		parent.replaceNodeWith(myReplacement);
	}

	protected TrieNode getNode(char location) {
		return subtries[indexFor(location)];
	}
	
	protected TrieNode[] getSubtries() {
		return this.subtries;
	}
	
	protected  void replaceNodeWith(TrieInnerNode newNode) {
		setNode(newNode, newNode.character);
	}
	
	private int indexFor(char location) {
		return characterSet.indexOf(location);
	}

	private TrieNode[] initialSubtries(Alphabet letters) {
		TrieNode[] emptyTries = new TrieNode[letters.size()];
		TrieNode leaf = new TrieNullNode(characterSet, this);
		Arrays.fill(emptyTries, leaf);
		return emptyTries;
	}

	private boolean isRootNode() {
		return character == NULL_CHAR;
	}

	private void setNode(TrieNode newNode, char location) {
		subtries[indexFor(location)] = newNode;
	}
}
